@@ -3,14 +3,14 @@ use std::hash::{Hash, Hasher};
 
 use pyo3::exceptions::PyIndexError;
 use pyo3::pyclass::CompareOp;
-use pyo3::types::{PyDict, PyIterator, PyNone, PyTuple, PyType};
-use pyo3::{exceptions::PyKeyError, types::PyMapping};
+use pyo3::types::{PyDict, PyIterator, PyTuple, PyType};
+use pyo3::{exceptions::PyKeyError, types::PyMapping, types::PyTupleMethods};
 use pyo3::{prelude::*, AsPyPointer, PyTypeInfo};
 use rpds::{
     HashTrieMap, HashTrieMapSync, HashTrieSet, HashTrieSetSync, List, ListSync, Queue, QueueSync,
 };
 
-#[derive(Clone, Debug)]
+#[derive(Debug)]
 struct Key {
     hash: isize,
     inner: PyObject,
@@ -35,6 +35,15 @@ impl PartialEq for Key {
     }
 }
 
+impl Key {
+    fn clone_ref(&self, py: Python<'_>) -> Self {
+        Key {
+            hash: self.hash,
+            inner: self.inner.clone_ref(py),
+        }
+    }
+}
+
 impl IntoPy<PyObject> for Key {
     fn into_py(self, py: Python<'_>) -> PyObject {
         self.inner.into_py(py)
@@ -48,10 +57,10 @@ unsafe impl AsPyPointer for Key {
 }
 
 impl<'source> FromPyObject<'source> for Key {
-    fn extract(ob: &'source PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<'source, PyAny>) -> PyResult<Self> {
         Ok(Key {
             hash: ob.hash()?,
-            inner: ob.into(),
+            inner: ob.clone().unbind(),
         })
     }
 }
@@ -69,7 +78,7 @@ impl From<HashTrieMapSync<Key, PyObject>> for HashTrieMapPy {
 }
 
 impl<'source> FromPyObject<'source> for HashTrieMapPy {
-    fn extract(ob: &'source PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<'source, PyAny>) -> PyResult<Self> {
         let mut ret = HashTrieMap::new_sync();
         if let Ok(mapping) = ob.downcast::<PyMapping>() {
             for each in mapping.items()?.iter()? {
@@ -89,8 +98,8 @@ impl<'source> FromPyObject<'source> for HashTrieMapPy {
 #[pymethods]
 impl HashTrieMapPy {
     #[new]
-    #[pyo3(signature = (value=None, **kwds))]
-    fn init(value: Option<HashTrieMapPy>, kwds: Option<&PyDict>) -> PyResult<Self> {
+    #[pyo3(signature = (value=None, ** kwds))]
+    fn init(value: Option<HashTrieMapPy>, kwds: Option<&Bound<'_, PyDict>>) -> PyResult<Self> {
         let mut map: HashTrieMapPy;
         if let Some(value) = value {
             map = value;
@@ -101,7 +110,7 @@ impl HashTrieMapPy {
         }
         if let Some(kwds) = kwds {
             for (k, v) in kwds {
-                map.inner.insert_mut(Key::extract(k)?, v.into());
+                map.inner.insert_mut(Key::extract_bound(&k)?, v.into());
             }
         }
         Ok(map)
@@ -117,9 +126,9 @@ impl HashTrieMapPy {
         }
     }
 
-    fn __getitem__(&self, key: Key) -> PyResult<PyObject> {
+    fn __getitem__(&self, key: Key, py: Python) -> PyResult<PyObject> {
         match self.inner.get(&key) {
-            Some(value) => Ok(value.to_owned()),
+            Some(value) => Ok(value.clone_ref(py)),
             None => Err(PyKeyError::new_err(key)),
         }
     }
@@ -154,7 +163,7 @@ impl HashTrieMapPy {
                     .inner
                     .iter()
                     .map(|(k1, v1)| (v1, other.inner.get(k1)))
-                    .map(|(v1, v2)| PyAny::eq(v1.extract(py)?, v2))
+                    .map(|(v1, v2)| v1.bind(py).eq(v2))
                     .all(|r| r.unwrap_or(false)))
             .into_py(py)),
             CompareOp::Ne => Ok((self.inner.size() != other.inner.size()
@@ -162,52 +171,58 @@ impl HashTrieMapPy {
                     .inner
                     .iter()
                     .map(|(k1, v1)| (v1, other.inner.get(k1)))
-                    .map(|(v1, v2)| PyAny::ne(v1.extract(py)?, v2))
+                    .map(|(v1, v2)| v1.bind(py).ne(v2))
                     .all(|r| r.unwrap_or(true)))
             .into_py(py)),
             _ => Ok(py.NotImplemented()),
         }
     }
 
-    fn __reduce__(slf: PyRef<Self>) -> (&PyType, (Vec<(Key, PyObject)>,)) {
+    fn __reduce__(slf: PyRef<Self>) -> (Bound<'_, PyType>, (Vec<(Key, PyObject)>,)) {
         (
-            HashTrieMapPy::type_object(slf.py()),
+            HashTrieMapPy::type_object_bound(slf.py()),
             (slf.inner
                 .iter()
-                .map(|(k, v)| (k.clone(), v.clone()))
+                .map(|(k, v)| (k.clone_ref(slf.py()), v.clone_ref(slf.py())))
                 .collect(),),
         )
     }
 
     #[classmethod]
-    fn convert(_cls: &PyType, value: &PyAny, py: Python) -> PyResult<PyObject> {
+    fn convert(
+        _cls: &Bound<'_, PyType>,
+        value: Bound<'_, PyAny>,
+        py: Python,
+    ) -> PyResult<PyObject> {
         if value.is_instance_of::<HashTrieMapPy>() {
-            Ok(value.into())
+            Ok(value.unbind())
         } else {
-            Ok(HashTrieMapPy::extract(value)?.into_py(py))
+            Ok(HashTrieMapPy::extract_bound(&value)?.into_py(py))
         }
     }
 
     #[classmethod]
+    #[pyo3(signature = (keys, val=None))]
     fn fromkeys(
-        _cls: &PyType,
-        keys: &PyAny,
-        val: Option<&PyAny>,
+        _cls: &Bound<'_, PyType>,
+        keys: &Bound<'_, PyAny>,
+        val: Option<&Bound<'_, PyAny>>,
         py: Python,
     ) -> PyResult<HashTrieMapPy> {
         let mut inner = HashTrieMap::new_sync();
-        let none = py.None();
-        let value = val.unwrap_or_else(|| none.as_ref(py));
+        let none = py.None().into_bound(py);
+        let value = val.unwrap_or(&none);
         for each in keys.iter()? {
-            let key = Key::extract(each?)?.to_owned();
-            inner.insert_mut(key, value.into());
+            let key = Key::extract_bound(&each?)?;
+            inner.insert_mut(key, value.clone().unbind());
         }
         Ok(HashTrieMapPy { inner })
     }
 
-    fn get(&self, key: Key, default: Option<PyObject>) -> Option<PyObject> {
+    #[pyo3(signature = (key, default=None))]
+    fn get(&self, key: Key, default: Option<PyObject>, py: Python) -> Option<PyObject> {
         if let Some(value) = self.inner.get(&key) {
-            Some(value.to_owned())
+            Some(value.clone_ref(py))
         } else {
             default
         }
@@ -242,9 +257,9 @@ impl HashTrieMapPy {
         }
     }
 
-    fn insert(&self, key: Key, value: &PyAny) -> HashTrieMapPy {
+    fn insert(&self, key: Key, value: Bound<'_, PyAny>) -> HashTrieMapPy {
         HashTrieMapPy {
-            inner: self.inner.insert(key, value.into()),
+            inner: self.inner.insert(key, value.unbind()),
         }
     }
 
@@ -258,17 +273,21 @@ impl HashTrieMapPy {
     }
 
     #[pyo3(signature = (*maps, **kwds))]
-    fn update(&self, maps: &PyTuple, kwds: Option<&PyDict>) -> PyResult<HashTrieMapPy> {
+    fn update(
+        &self,
+        maps: &Bound<'_, PyTuple>,
+        kwds: Option<&Bound<'_, PyDict>>,
+    ) -> PyResult<HashTrieMapPy> {
         let mut inner = self.inner.clone();
         for value in maps {
-            let map = HashTrieMapPy::extract(value)?;
+            let map = HashTrieMapPy::extract_bound(&value)?;
             for (k, v) in &map.inner {
-                inner.insert_mut(k.to_owned(), v.to_owned());
+                inner.insert_mut(k.clone_ref(value.py()), v.clone_ref(value.py()));
             }
         }
         if let Some(kwds) = kwds {
             for (k, v) in kwds {
-                inner.insert_mut(Key::extract(k)?, v.extract()?);
+                inner.insert_mut(Key::extract_bound(&k)?, v.extract()?);
             }
         }
         Ok(HashTrieMapPy { inner })
@@ -287,7 +306,7 @@ impl KeysIterator {
     }
 
     fn __next__(mut slf: PyRefMut<'_, Self>) -> Option<Key> {
-        let first = slf.inner.keys().next()?.to_owned();
+        let first = slf.inner.keys().next()?.clone_ref(slf.py());
         slf.inner = slf.inner.remove(&first);
         Some(first)
     }
@@ -306,7 +325,7 @@ impl ValuesIterator {
 
     fn __next__(mut slf: PyRefMut<'_, Self>) -> Option<PyObject> {
         let kv = slf.inner.iter().next()?;
-        let value = kv.1.to_owned();
+        let value = kv.1.clone_ref(slf.py());
         slf.inner = slf.inner.remove(kv.0);
         Some(value)
     }
@@ -325,9 +344,11 @@ impl ItemsIterator {
 
     fn __next__(mut slf: PyRefMut<'_, Self>) -> Option<(Key, PyObject)> {
         let kv = slf.inner.iter().next()?;
-        let key = kv.0.to_owned();
-        let value = kv.1.to_owned();
+        let key = kv.0.clone_ref(slf.py());
+        let value = kv.1.clone_ref(slf.py());
+
         slf.inner = slf.inner.remove(kv.0);
+
         Some((key, value))
     }
 }
@@ -343,65 +364,67 @@ impl KeysView {
         self.inner.contains_key(&key)
     }
 
-    fn __eq__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? != slf.inner.size() {
+    fn __eq__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? != slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
-            if !slf.inner.contains_key(&Key::extract(each?)?) {
+            if !slf.inner.contains_key(&Key::extract_bound(&each?)?) {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __lt__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? <= slf.inner.size() {
+    fn __lt__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? <= slf.inner.size() {
             return Ok(false);
         }
+
         for each in slf.inner.keys() {
-            if !other.contains(each.inner.to_owned())? {
+            if !other.contains(each.inner.clone_ref(slf.py()))? {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __le__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? < slf.inner.size() {
+    fn __le__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? < slf.inner.size() {
             return Ok(false);
         }
+
         for each in slf.inner.keys() {
-            if !other.contains(each.inner.to_owned())? {
+            if !other.contains(each.inner.clone_ref(slf.py()))? {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __gt__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? >= slf.inner.size() {
+    fn __gt__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? >= slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
-            if !slf.inner.contains_key(&Key::extract(each?)?) {
+            if !slf.inner.contains_key(&Key::extract_bound(&each?)?) {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __ge__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? > slf.inner.size() {
+    fn __ge__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? > slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
-            if !slf.inner.contains_key(&Key::extract(each?)?) {
+            if !slf.inner.contains_key(&Key::extract_bound(&each?)?) {
                 return Ok(false);
             }
         }
@@ -418,17 +441,17 @@ impl KeysView {
         slf.inner.size()
     }
 
-    fn __and__(slf: PyRef<'_, Self>, other: &PyAny) -> PyResult<HashTrieSetPy> {
+    fn __and__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>) -> PyResult<HashTrieSetPy> {
         KeysView::intersection(slf, other)
     }
 
-    fn __or__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<KeysView> {
+    fn __or__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<KeysView> {
         KeysView::union(slf, other, py)
     }
 
     fn __repr__(&self, py: Python) -> String {
         let contents = self.inner.into_iter().map(|(k, _)| {
-            k.clone()
+            k.clone_ref(py)
                 .inner
                 .into_py(py)
                 .call_method0(py, "__repr__")
@@ -438,11 +461,11 @@ impl KeysView {
         format!("keys_view({{{}}})", contents.collect::<Vec<_>>().join(", "))
     }
 
-    fn intersection(slf: PyRef<'_, Self>, other: &PyAny) -> PyResult<HashTrieSetPy> {
+    fn intersection(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>) -> PyResult<HashTrieSetPy> {
         // TODO: iterate over the shorter one if it's got a length
         let mut inner = HashTrieSet::new_sync();
         for each in other.iter()? {
-            let key = Key::extract(each?)?;
+            let key = Key::extract_bound(&each?)?;
             if slf.inner.contains_key(&key) {
                 inner.insert_mut(key);
             }
@@ -450,13 +473,12 @@ impl KeysView {
         Ok(HashTrieSetPy { inner })
     }
 
-    fn union(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<KeysView> {
+    fn union(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<KeysView> {
         // There doesn't seem to be a low-effort way to get a HashTrieSet out of a map,
         // so we just keep our map and add values we'll ignore.
         let mut inner = slf.inner.clone();
-        let none = PyNone::get(py);
         for each in other.iter()? {
-            inner.insert_mut(Key::extract(each?)?, none.into());
+            inner.insert_mut(Key::extract_bound(&each?)?, py.None());
         }
         Ok(KeysView { inner })
     }
@@ -495,12 +517,16 @@ struct ItemsView {
     inner: HashTrieMapSync<Key, PyObject>,
 }
 
+#[derive(FromPyObject)]
+struct ItemViewQuery(Key, PyObject);
+
 #[pymethods]
 impl ItemsView {
-    fn __contains__(slf: PyRef<'_, Self>, item: (Key, &PyAny)) -> PyResult<bool> {
+    fn __contains__(slf: PyRef<'_, Self>, item: ItemViewQuery) -> PyResult<bool> {
         if let Some(value) = slf.inner.get(&item.0) {
-            return item.1.eq(value);
+            return item.1.bind(slf.py()).eq(value);
         }
+
         Ok(false)
     }
 
@@ -514,13 +540,13 @@ impl ItemsView {
         slf.inner.size()
     }
 
-    fn __eq__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? != slf.inner.size() {
+    fn __eq__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? != slf.inner.size() {
             return Ok(false);
         }
         for (k, v) in slf.inner.iter() {
-            if !other.contains((k.inner.to_owned(), v))? {
+            if !other.contains((k.inner.clone_ref(slf.py()), v))? {
                 return Ok(false);
             }
         }
@@ -529,19 +555,19 @@ impl ItemsView {
 
     fn __repr__(&self, py: Python) -> String {
         let contents = self.inner.into_iter().map(|(k, v)| {
-            let tuple = PyTuple::new(py, [k.inner.to_owned(), v.to_owned()]);
+            let tuple = PyTuple::new_bound(py, [k.inner.clone_ref(py), v.clone_ref(py)]);
             format!("{:?}", tuple)
         });
         format!("items_view([{}])", contents.collect::<Vec<_>>().join(", "))
     }
 
-    fn __lt__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? <= slf.inner.size() {
+    fn __lt__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? <= slf.inner.size() {
             return Ok(false);
         }
         for (k, v) in slf.inner.iter() {
-            let pair = PyTuple::new(py, [k.inner.to_owned(), v.to_owned()]);
+            let pair = PyTuple::new_bound(py, [k.inner.clone_ref(py), v.clone_ref(py)]);
             // FIXME: needs to compare
             if !other.contains(pair)? {
                 return Ok(false);
@@ -550,13 +576,13 @@ impl ItemsView {
         Ok(true)
     }
 
-    fn __le__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? < slf.inner.size() {
+    fn __le__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? < slf.inner.size() {
             return Ok(false);
         }
         for (k, v) in slf.inner.iter() {
-            let pair = PyTuple::new(py, [k.inner.to_owned(), v.to_owned()]);
+            let pair = PyTuple::new_bound(py, [k.inner.clone_ref(py), v.clone_ref(py)]);
             // FIXME: needs to compare
             if !other.contains(pair)? {
                 return Ok(false);
@@ -565,17 +591,17 @@ impl ItemsView {
         Ok(true)
     }
 
-    fn __gt__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? >= slf.inner.size() {
+    fn __gt__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? >= slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
             let kv = each?;
             let k = kv.get_item(0)?;
-            match slf.inner.get(&Key::extract(k)?) {
+            match slf.inner.get(&Key::extract_bound(&k)?) {
                 Some(value) => {
-                    let pair = PyTuple::new(py, [k, value.as_ref(py)]);
+                    let pair = PyTuple::new_bound(py, [k, value.bind(py).clone()]);
                     if !pair.eq(kv)? {
                         return Ok(false);
                     }
@@ -586,17 +612,17 @@ impl ItemsView {
         Ok(true)
     }
 
-    fn __ge__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? > slf.inner.size() {
+    fn __ge__(slf: PyRef<'_, Self>, other: &Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? > slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
             let kv = each?;
             let k = kv.get_item(0)?;
-            match slf.inner.get(&Key::extract(k)?) {
+            match slf.inner.get(&Key::extract_bound(&k)?) {
                 Some(value) => {
-                    let pair = PyTuple::new(py, [k, value.as_ref(py)]);
+                    let pair = PyTuple::new_bound(py, [k, value.bind(py).clone()]);
                     if !pair.eq(kv)? {
                         return Ok(false);
                     }
@@ -607,39 +633,55 @@ impl ItemsView {
         Ok(true)
     }
 
-    fn __and__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<HashTrieSetPy> {
+    fn __and__(
+        slf: PyRef<'_, Self>,
+        other: &Bound<'_, PyAny>,
+        py: Python,
+    ) -> PyResult<HashTrieSetPy> {
         ItemsView::intersection(slf, other, py)
     }
 
-    fn __or__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<HashTrieSetPy> {
+    fn __or__(
+        slf: PyRef<'_, Self>,
+        other: &Bound<'_, PyAny>,
+        py: Python,
+    ) -> PyResult<HashTrieSetPy> {
         ItemsView::union(slf, other, py)
     }
 
-    fn intersection(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<HashTrieSetPy> {
+    fn intersection(
+        slf: PyRef<'_, Self>,
+        other: &Bound<'_, PyAny>,
+        py: Python,
+    ) -> PyResult<HashTrieSetPy> {
         // TODO: iterate over the shorter one if it's got a length
         let mut inner = HashTrieSet::new_sync();
         for each in other.iter()? {
             let kv = each?;
             let k = kv.get_item(0)?;
-            if let Some(value) = slf.inner.get(&Key::extract(k)?) {
-                let pair = PyTuple::new(py, [k, value.as_ref(py)]);
+            if let Some(value) = slf.inner.get(&Key::extract_bound(&k)?) {
+                let pair = PyTuple::new_bound(py, [k, value.bind(py).clone()]);
                 if pair.eq(kv)? {
-                    inner.insert_mut(Key::extract(pair)?);
+                    inner.insert_mut(Key::extract_bound(&pair)?);
                 }
             }
         }
         Ok(HashTrieSetPy { inner })
     }
 
-    fn union(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<HashTrieSetPy> {
+    fn union(
+        slf: PyRef<'_, Self>,
+        other: &Bound<'_, PyAny>,
+        py: Python,
+    ) -> PyResult<HashTrieSetPy> {
         // TODO: this is very inefficient, but again can't seem to get a HashTrieSet out of ourself
         let mut inner = HashTrieSet::new_sync();
         for (k, v) in slf.inner.iter() {
-            let pair = PyTuple::new(py, [k.inner.to_owned(), v.to_owned()]);
-            inner.insert_mut(Key::extract(pair)?);
+            let pair = PyTuple::new_bound(py, [k.inner.clone_ref(py), v.clone_ref(py)]);
+            inner.insert_mut(Key::extract_bound(&pair)?);
         }
         for each in other.iter()? {
-            inner.insert_mut(Key::extract(each?)?);
+            inner.insert_mut(Key::extract_bound(&each?)?);
         }
         Ok(HashTrieSetPy { inner })
     }
@@ -652,7 +694,7 @@ struct HashTrieSetPy {
 }
 
 impl<'source> FromPyObject<'source> for HashTrieSetPy {
-    fn extract(ob: &'source PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<'source, PyAny>) -> PyResult<Self> {
         let mut ret = HashTrieSet::new_sync();
         for each in ob.iter()? {
             let k: Key = each?.extract()?;
@@ -665,6 +707,7 @@ impl<'source> FromPyObject<'source> for HashTrieSetPy {
 #[pymethods]
 impl HashTrieSetPy {
     #[new]
+    #[pyo3(signature = (value=None))]
     fn init(value: Option<HashTrieSetPy>) -> Self {
         if let Some(value) = value {
             value
@@ -679,20 +722,20 @@ impl HashTrieSetPy {
         self.inner.contains(&key)
     }
 
-    fn __and__(&self, other: &Self) -> Self {
-        self.intersection(other)
+    fn __and__(&self, other: &Self, py: Python) -> Self {
+        self.intersection(other, py)
     }
 
-    fn __or__(&self, other: &Self) -> Self {
-        self.union(other)
+    fn __or__(&self, other: &Self, py: Python) -> Self {
+        self.union(other, py)
     }
 
     fn __sub__(&self, other: &Self) -> Self {
         self.difference(other)
     }
 
-    fn __xor__(&self, other: &Self) -> Self {
-        self.symmetric_difference(other)
+    fn __xor__(&self, other: &Self, py: Python) -> Self {
+        self.symmetric_difference(other, py)
     }
 
     fn __iter__(slf: PyRef<'_, Self>) -> SetIterator {
@@ -707,7 +750,7 @@ impl HashTrieSetPy {
 
     fn __repr__(&self, py: Python) -> String {
         let contents = self.inner.into_iter().map(|k| {
-            k.clone()
+            k.clone_ref(py)
                 .into_py(py)
                 .call_method0(py, "__repr__")
                 .and_then(|r| r.extract(py))
@@ -719,75 +762,75 @@ impl HashTrieSetPy {
         )
     }
 
-    fn __eq__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? != slf.inner.size() {
+    fn __eq__(slf: PyRef<'_, Self>, other: Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? != slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
-            if !slf.inner.contains(&Key::extract(each?)?) {
+            if !slf.inner.contains(&Key::extract_bound(&each?)?) {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __lt__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? <= slf.inner.size() {
+    fn __lt__(slf: PyRef<'_, Self>, other: Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? <= slf.inner.size() {
             return Ok(false);
         }
         for each in slf.inner.iter() {
-            if !other.contains(each.inner.to_owned())? {
+            if !other.contains(each.inner.clone_ref(py))? {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __le__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? < slf.inner.size() {
+    fn __le__(slf: PyRef<'_, Self>, other: Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? < slf.inner.size() {
             return Ok(false);
         }
         for each in slf.inner.iter() {
-            if !other.contains(each.inner.to_owned())? {
+            if !other.contains(each.inner.clone_ref(slf.py()))? {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __gt__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? >= slf.inner.size() {
+    fn __gt__(slf: PyRef<'_, Self>, other: Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? >= slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
-            if !slf.inner.contains(&Key::extract(each?)?) {
+            if !slf.inner.contains(&Key::extract_bound(&each?)?) {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __ge__(slf: PyRef<'_, Self>, other: &PyAny, py: Python) -> PyResult<bool> {
-        let abc = PyModule::import(py, "collections.abc")?;
-        if !other.is_instance(abc.getattr("Set")?)? || other.len()? > slf.inner.size() {
+    fn __ge__(slf: PyRef<'_, Self>, other: Bound<'_, PyAny>, py: Python) -> PyResult<bool> {
+        let abc = PyModule::import_bound(py, "collections.abc")?;
+        if !other.is_instance(&abc.getattr("Set")?)? || other.len()? > slf.inner.size() {
             return Ok(false);
         }
         for each in other.iter()? {
-            if !slf.inner.contains(&Key::extract(each?)?) {
+            if !slf.inner.contains(&Key::extract_bound(&each?)?) {
                 return Ok(false);
             }
         }
         Ok(true)
     }
 
-    fn __reduce__(slf: PyRef<Self>) -> (&PyType, (Vec<Key>,)) {
+    fn __reduce__(slf: PyRef<Self>) -> (Bound<'_, PyType>, (Vec<Key>,)) {
         (
-            HashTrieSetPy::type_object(slf.py()),
-            (slf.inner.iter().cloned().collect(),),
+            HashTrieSetPy::type_object_bound(slf.py()),
+            (slf.inner.iter().map(|e| e.clone_ref(slf.py())).collect(),),
         )
     }
 
@@ -825,7 +868,7 @@ impl HashTrieSetPy {
         HashTrieSetPy { inner }
     }
 
-    fn intersection(&self, other: &Self) -> HashTrieSetPy {
+    fn intersection(&self, other: &Self, py: Python) -> HashTrieSetPy {
         let mut inner: HashTrieSetSync<Key> = HashTrieSet::new_sync();
         let larger: &HashTrieSetSync<Key>;
         let iter;
@@ -838,13 +881,13 @@ impl HashTrieSetPy {
         }
         for value in iter {
             if larger.contains(value) {
-                inner.insert_mut(value.to_owned());
+                inner.insert_mut(value.clone_ref(py));
             }
         }
         HashTrieSetPy { inner }
     }
 
-    fn symmetric_difference(&self, other: &Self) -> HashTrieSetPy {
+    fn symmetric_difference(&self, other: &Self, py: Python) -> HashTrieSetPy {
         let mut inner: HashTrieSetSync<Key>;
         let iter;
         if self.inner.size() > other.inner.size() {
@@ -858,13 +901,13 @@ impl HashTrieSetPy {
             if inner.contains(value) {
                 inner.remove_mut(value);
             } else {
-                inner.insert_mut(value.to_owned());
+                inner.insert_mut(value.clone_ref(py));
             }
         }
         HashTrieSetPy { inner }
     }
 
-    fn union(&self, other: &Self) -> HashTrieSetPy {
+    fn union(&self, other: &Self, py: Python) -> HashTrieSetPy {
         let mut inner: HashTrieSetSync<Key>;
         let iter;
         if self.inner.size() > other.inner.size() {
@@ -875,18 +918,18 @@ impl HashTrieSetPy {
             iter = self.inner.iter();
         }
         for value in iter {
-            inner.insert_mut(value.to_owned());
+            inner.insert_mut(value.clone_ref(py));
         }
         HashTrieSetPy { inner }
     }
 
     #[pyo3(signature = (*iterables))]
-    fn update(&self, iterables: &PyTuple) -> PyResult<HashTrieSetPy> {
+    fn update(&self, iterables: Bound<'_, PyTuple>) -> PyResult<HashTrieSetPy> {
         let mut inner = self.inner.clone();
         for each in iterables {
             let iter = each.iter()?;
             for value in iter {
-                inner.insert_mut(Key::extract(value?)?.to_owned());
+                inner.insert_mut(Key::extract_bound(&value?)?);
             }
         }
         Ok(HashTrieSetPy { inner })
@@ -905,7 +948,7 @@ impl SetIterator {
     }
 
     fn __next__(mut slf: PyRefMut<'_, Self>) -> Option<Key> {
-        let first = slf.inner.iter().next()?.to_owned();
+        let first = slf.inner.iter().next()?.clone_ref(slf.py());
         slf.inner = slf.inner.remove(&first);
         Some(first)
     }
@@ -924,10 +967,10 @@ impl From<ListSync<PyObject>> for ListPy {
 }
 
 impl<'source> FromPyObject<'source> for ListPy {
-    fn extract(ob: &'source PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<'source, PyAny>) -> PyResult<Self> {
         let mut ret = List::new_sync();
-        let reversed = PyModule::import(ob.py(), "builtins")?.getattr("reversed")?;
-        let rob: &PyIterator = reversed.call1((ob,))?.iter()?;
+        let reversed = PyModule::import_bound(ob.py(), "builtins")?.getattr("reversed")?;
+        let rob: Bound<'_, PyIterator> = reversed.call1((ob,))?.iter()?;
         for each in rob {
             ret.push_front_mut(each?.extract()?);
         }
@@ -939,7 +982,7 @@ impl<'source> FromPyObject<'source> for ListPy {
 impl ListPy {
     #[new]
     #[pyo3(signature = (*elements))]
-    fn init(elements: &PyTuple) -> PyResult<Self> {
+    fn init(elements: &Bound<'_, PyTuple>) -> PyResult<Self> {
         let mut ret: ListPy;
         if elements.len() == 1 {
             ret = elements.get_item(0)?.extract()?;
@@ -978,7 +1021,7 @@ impl ListPy {
                     .inner
                     .iter()
                     .zip(other.inner.iter())
-                    .map(|(e1, e2)| PyAny::eq(e1.extract(py)?, e2))
+                    .map(|(e1, e2)| e1.bind(py).eq(e2))
                     .all(|r| r.unwrap_or(false)))
             .into_py(py)),
             CompareOp::Ne => Ok((self.inner.len() != other.inner.len()
@@ -986,7 +1029,7 @@ impl ListPy {
                     .inner
                     .iter()
                     .zip(other.inner.iter())
-                    .map(|(e1, e2)| PyAny::ne(e1.extract(py)?, e2))
+                    .map(|(e1, e2)| e1.bind(py).ne(e2))
                     .any(|r| r.unwrap_or(true)))
             .into_py(py)),
             _ => Ok(py.NotImplemented()),
@@ -1005,10 +1048,10 @@ impl ListPy {
         }
     }
 
-    fn __reduce__(slf: PyRef<Self>) -> (&PyType, (Vec<PyObject>,)) {
+    fn __reduce__(slf: PyRef<Self>) -> (Bound<'_, PyType>, (Vec<PyObject>,)) {
         (
-            ListPy::type_object(slf.py()),
-            (slf.inner.iter().cloned().collect(),),
+            ListPy::type_object_bound(slf.py()),
+            (slf.inner.iter().map(|e| e.clone_ref(slf.py())).collect(),),
         )
     }
 
@@ -1053,8 +1096,11 @@ impl ListIterator {
     }
 
     fn __next__(mut slf: PyRefMut<'_, Self>) -> Option<PyObject> {
-        let first = slf.inner.first()?.to_owned();
+        let first_op = slf.inner.first()?;
+        let first = first_op.clone_ref(slf.py());
+
         slf.inner = slf.inner.drop_first()?;
+
         Some(first)
     }
 }
@@ -1071,7 +1117,8 @@ impl QueueIterator {
     }
 
     fn __next__(mut slf: PyRefMut<'_, Self>) -> Option<PyObject> {
-        let first = slf.inner.peek()?.to_owned();
+        let first_op = slf.inner.peek()?;
+        let first = first_op.clone_ref(slf.py());
         slf.inner = slf.inner.dequeue()?;
         Some(first)
     }
@@ -1090,7 +1137,7 @@ impl From<QueueSync<PyObject>> for QueuePy {
 }
 
 impl<'source> FromPyObject<'source> for QueuePy {
-    fn extract(ob: &'source PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<'source, PyAny>) -> PyResult<Self> {
         let mut ret = Queue::new_sync();
         for each in ob.iter()? {
             ret.enqueue_mut(each?.extract()?);
@@ -1103,7 +1150,7 @@ impl<'source> FromPyObject<'source> for QueuePy {
 impl QueuePy {
     #[new]
     #[pyo3(signature = (*elements))]
-    fn init(elements: &PyTuple, py: Python<'_>) -> PyResult<Self> {
+    fn init(elements: &Bound<'_, PyTuple>, py: Python<'_>) -> PyResult<Self> {
         let mut ret: QueuePy;
         if elements.len() == 1 {
             ret = elements.get_item(0)?.extract()?;
@@ -1126,12 +1173,12 @@ impl QueuePy {
                 .inner
                 .iter()
                 .zip(other.inner.iter())
-                .map(|(e1, e2)| PyAny::eq(e1.extract(py)?, e2))
+                .map(|(e1, e2)| e1.bind(py).eq(e2))
                 .all(|r| r.unwrap_or(false))
     }
 
     fn __hash__(&self, py: Python<'_>) -> PyResult<u64> {
-        let hash = PyModule::import(py, "builtins")?.getattr("hash")?;
+        let hash = PyModule::import_bound(py, "builtins")?.getattr("hash")?;
         let mut hasher = DefaultHasher::new();
         for each in &self.inner {
             let n: i64 = hash.call1((each.into_py(py),))?.extract()?;
@@ -1146,7 +1193,7 @@ impl QueuePy {
                 .inner
                 .iter()
                 .zip(other.inner.iter())
-                .map(|(e1, e2)| PyAny::ne(e1.extract(py)?, e2))
+                .map(|(e1, e2)| e1.bind(py).ne(e2))
                 .any(|r| r.unwrap_or(true))
     }
 
@@ -1162,8 +1209,7 @@ impl QueuePy {
 
     fn __repr__(&self, py: Python) -> String {
         let contents = self.inner.into_iter().map(|k| {
-            k.clone()
-                .into_py(py)
+            k.into_py(py)
                 .call_method0(py, "__repr__")
                 .and_then(|r| r.extract(py))
                 .unwrap_or("<repr failed>".to_owned())
@@ -1172,9 +1218,9 @@ impl QueuePy {
     }
 
     #[getter]
-    fn peek(&self) -> PyResult<PyObject> {
+    fn peek(&self, py: Python) -> PyResult<PyObject> {
         if let Some(peeked) = self.inner.peek() {
-            Ok(peeked.to_owned())
+            Ok(peeked.clone_ref(py))
         } else {
             Err(PyIndexError::new_err("peeked an empty queue"))
         }
@@ -1185,7 +1231,7 @@ impl QueuePy {
         self.inner.is_empty()
     }
 
-    fn enqueue(&self, value: &PyAny) -> Self {
+    fn enqueue(&self, value: Bound<'_, PyAny>) -> Self {
         QueuePy {
             inner: self.inner.enqueue(value.into()),
         }
@@ -1202,7 +1248,7 @@ impl QueuePy {
 
 #[pymodule]
 #[pyo3(name = "rpds")]
-fn rpds_py(py: Python, m: &PyModule) -> PyResult<()> {
+fn rpds_py(py: Python, m: &Bound<'_, PyModule>) -> PyResult<()> {
     m.add_class::<HashTrieMapPy>()?;
     m.add_class::<HashTrieSetPy>()?;
     m.add_class::<ListPy>()?;
@@ -1210,24 +1256,24 @@ fn rpds_py(py: Python, m: &PyModule) -> PyResult<()> {
 
     PyMapping::register::<HashTrieMapPy>(py)?;
 
-    let abc = PyModule::import(py, "collections.abc")?;
+    let abc = PyModule::import_bound(py, "collections.abc")?;
 
     abc.getattr("Set")?
-        .call_method1("register", (HashTrieSetPy::type_object(py),))?;
+        .call_method1("register", (HashTrieSetPy::type_object_bound(py),))?;
 
     abc.getattr("MappingView")?
-        .call_method1("register", (KeysView::type_object(py),))?;
+        .call_method1("register", (KeysView::type_object_bound(py),))?;
     abc.getattr("MappingView")?
-        .call_method1("register", (ValuesView::type_object(py),))?;
+        .call_method1("register", (ValuesView::type_object_bound(py),))?;
     abc.getattr("MappingView")?
-        .call_method1("register", (ItemsView::type_object(py),))?;
+        .call_method1("register", (ItemsView::type_object_bound(py),))?;
 
     abc.getattr("KeysView")?
-        .call_method1("register", (KeysView::type_object(py),))?;
+        .call_method1("register", (KeysView::type_object_bound(py),))?;
     abc.getattr("ValuesView")?
-        .call_method1("register", (ValuesView::type_object(py),))?;
+        .call_method1("register", (ValuesView::type_object_bound(py),))?;
     abc.getattr("ItemsView")?
-        .call_method1("register", (ItemsView::type_object(py),))?;
+        .call_method1("register", (ItemsView::type_object_bound(py),))?;
 
     Ok(())
 }
